@@ -1,5 +1,5 @@
 
-# Copyright (c) 2001-2003 Nathan Wiger <nate@sun.com>
+# Copyright (c) 2001-2004 Nathan Wiger <nate@sun.com>
 # Please visit www.formbuilder.org for support and examples
 # Use "perldoc FormBuilder.pod" for documentation
 
@@ -8,7 +8,7 @@ package CGI::FormBuilder;
 use Carp;
 use strict;
 use vars qw($VERSION $CGIMOD $CGI $AUTOLOAD);
-$VERSION = do { my @r=(q$Revision: 2.11 $=~/\d+/g); sprintf "%d."."%02d"x$#r,@r };
+$VERSION = do { my @r=(q$Revision: 2.13 $=~/\d+/g); sprintf "%d."."%02d"x$#r,@r };
 
 # We used to try to use CGI::Minimal for better speed, but
 # unfortunately its support for file uploads it unsuitable
@@ -35,7 +35,8 @@ my %VALID = (
     FLOAT => '/^-?\s*[0-9]+\.[0-9]+$/',
     PHONE => ['/^\d{3}\-\d{3}\-\d{4}$|^\(\d{3}\)\s+\d{3}\-\d{4}$/', '123-456-7890'],
     INTPHONE => ['/^\+\d+[\s\-][\d\-\s]+$/', '+prefix local-number'],
-    EMAIL => ['/^[\w\-\+\.]+\@[a-zA-Z0-9][-a-zA-Z0-9\.]*\.[a-zA-Z]+$/', 'name@host.domain'],
+    ALLPHONE => ['/^(\d{3}[\-\.]?\s*|\(\d{3}\)\s*)\d{3}[\-\.]?\s*\d{4}$/', '(123) 456-7890'],
+    EMAIL => ['/^[\w\-\+\._]+\@[a-zA-Z0-9][-a-zA-Z0-9\.]*\.[a-zA-Z]+$/', 'name@host.domain'],
     CARD  => '/^\d{4}[\- ]?\d{4}[\- ]?\d{4}[\- ]?\d{4}$|^\d{4}[\- ]?\d{6}[\- ]?\d{5}$/',
     MMYY  => ['/^(0?[1-9]|1[0-2])\/?[0-9]{2}$/', 'MM/YY'],
     MMYYYY=> ['/^(0?[1-9]|1[0-2])\/?[0-9]{4}$/', 'MM/YYYY'],
@@ -47,13 +48,13 @@ my %VALID = (
     COUNTRY => ['/^[a-zA-Z]{2}$/', 'two-letter abbr'],
     IPV4  => ['/^([0-1]??\d{1,2}|2[0-4]\d|25[0-5])\.([0-1]??\d{1,2}|2[0-4]\d|25[0-5])\.([0-1]??\d{1,2}|2[0-4]\d|25[0-5])\.([0-1]??\d{1,2}|2[0-4]\d|25[0-5])$/', 'IP address'],
     NETMASK => ['/^([0-1]??\d{1,2}|2[0-4]\d|25[0-5])\.([0-1]??\d{1,2}|2[0-4]\d|25[0-5])\.([0-1]??\d{1,2}|2[0-4]\d|25[0-5])\.([0-1]??\d{1,2}|2[0-4]\d|25[0-5])$/', 'IP netmask'],
-    FILE  => ['/^[\/\w\.\-]+$/', 'UNIX format'],
-    WINFILE => ['/^[a-zA-Z]:\\[\\\w\s\.\-]+$/', 'Windows format'],
-    MACFILE => ['/^[:\w\.\-]+$/', 'Mac format'],
-    USER  => ['/^[-a-zA-Z0-9]{4,8}$/', '4-8 characters'],  # require a 4-8 char username
+    FILE  => ['/(^[\w\+_\040\#\(\)\{\}\[\]\/\-\^,\.:;&%@\\~]+\$?$)/ ', 'UNIX format'],
+    WINFILE => ['/^(([a-zA-Z]:)?\\?[\w\+_\040\(\)\{\}\[\]\/\-\^,\.;&%@\$!\\~\#]+)$/', 'Windows format'],
+    MACFILE => ['/^[:\w\.\-_]+$/', 'Mac format'],
+    USER  => ['/^[-a-zA-Z0-9_]{4,8}$/', '4-8 characters'],  # require a 4-8 char username
     HOST  => ['/^[a-zA-Z0-9][-a-zA-Z0-9]*$/', 'valid hostname'],
     DOMAIN=> ['/^[a-zA-Z0-9][-a-zA-Z0-9\.]*\.[a-zA-Z]+$/', 'DNS domain'],   # mostly correct, but allows "dom.c-o.uk"
-    ETHER => ['/^[\da-f]{1,2}[\.:]?[\da-f]{1,2}[\.:]?[\da-f]{1,2}[\.:]?[\da-f]{1,2}[\.:]?[\da-f]{1,2}[\.:]?[\da-f]{1,2}$/i', 'ethernet' ],
+    ETHER => ['/^[\dA-Fa-f]{1,2}[.:][\dA-Fa-f]{1,2}[.:][\dA-Fa-f]{1,2}[.:][\dA-Fa-f]{1,2}[.:][\dA-Fa-f]{1,2}[.:][\dA-Fa-f]{1,2}$|^[\dA-Fa-f]{12}$/', 'ethernet' ],
     # Many thanks to Mark Belanger for these additions
     FNAME => '/^[a-zA-Z]+[- ]?[a-zA-Z]*$/',
     LNAME => '/^[a-zA-Z]+[- ]?[a-zA-Z]+\s*,?([a-zA-Z]+|[a-zA-Z]+\.)?$/',
@@ -69,10 +70,10 @@ my %VALID = (
 my @OURATTR = qw(
     attr body checknum comment debug delete fieldattr fields fieldtype font
     force header invalid javascript keepextras label labels lalign
-    linebreaks multiple nameopts options override params radionum required
+    linebreaks nameopts options override params radionum required
     reset selectnum smartness sortopts static sticky submit table
     template title type_orig validate valign value_orig values messages 
-    columns
+    columns ulist
 );
 
 # trick for speedy lookup
@@ -389,7 +390,7 @@ sub _initfields () {
                 # the options are the names of the US states + DC (51)
                 $self->{fields}{$field}{options} ||= $OPTIONS{STATE};
                 debug 2, "via 'smartness' auto-determined options for '$field' field";
-            } elsif ($field =~ /^countr/i) {
+            } elsif ($field =~ /^countr/i) {    # "country" or "countries"
                 $self->{opt}{validate}{$field} ||= 'COUNTRY';
                 # the options are the currently-valid country codes, US first, of course :-)
                 $self->{fields}{$field}{options} ||= $OPTIONS{COUNTRY};
@@ -450,13 +451,16 @@ sub _tag ($;@) {
         # element, returning everything else (for an html tag)
         my $key = shift;
         # Eeek, I used "text" here and body takes a text attr!!
-        if ($OURATTR{$key} || ($key eq 'text' && $name ne 'body')) {
+        if ($OURATTR{$key} || ($key eq 'text' && $name ne 'body')
+                           || ($key eq 'multiple' && $name ne 'select')) {
             shift; next;
         }
         my $val = _escapehtml shift;    # minimalist HTML escaping
         push @tag, qq($key="$val");
     }
-    return '<' . join(' ', $name, sort @tag) . '>';
+    my $htag = join(' ', $name, sort @tag);
+    $htag .= ' /' if $name eq 'input';  # XHTML self-closing
+    return '<' . $htag . '>';
 }
 
 sub _expreqd ($$) {
@@ -498,8 +502,8 @@ sub new () {
     $args{method} ||= $method if $method;
 
     # Warning to try and catch user error
-    belch "You won't be able to get at any form values unless you specify 'fields' to new()"
-        unless $args{fields};
+    #belch "You won't be able to get at any form values unless you specify 'fields' to new()"
+        #unless $args{fields};
 
     $DEBUG = delete $args{debug} || 0;   # recall that delete returns the val deleted
 
@@ -637,7 +641,20 @@ sub field () {
             return;
         } elsif ($k eq 'validate') {
             # per-field validation; can clear with validate => undef
-            $self->{opt}{validate}{"$args{name}"} = $v;
+            $self->{opt}{$k}{"$args{name}"} = $v;
+        } elsif ($k eq 'required') {
+            # per-field validation; can clear with validate => undef
+            if (exists $self->{opt}{$k}) {
+                if (ref $self->{opt}{$k}) {
+                    push @{$self->{opt}{$k}}, $args{name};
+                } elsif ($self->{opt}{$k}) {
+                    belch "Cannot override required option in field() if set to 'ALL' or 'NONE'";
+                }
+            } elsif ($v) {
+                push @{$self->{opt}{$k}}, $args{name};
+            } else {
+                belch "Setting required => 0 in field() does not currently work right";
+            }
         }
         $self->{fields}{"$args{name}"}{$k} = $v;
         #debug 2, "\$self->{fields}{$args{name}}{$k} = $v";
@@ -772,7 +789,7 @@ sub render () {
 
     # How to handle line breaks - include <br> only if not a table
     my $br = $args{linebreaks}
-                ? ($args{table} ? "\n" : "<br>\n") : '';
+                ? ($args{table} ? "\n" : "<br />\n") : '';
 
     # For holding the JavaScript validation code
     my $jsfunc = '';
@@ -781,7 +798,7 @@ sub render () {
     # User-specified jsfunc options
     my $ajsf = delete $args{jsfunc} || '';
 
-    if ($args{javascript} && $args{validate} || $args{required}) {
+    if ($args{javascript} && ($args{validate} || $args{required})) {
         $jsfunc  = "function $jsname (form) {\n"
                  . "    var alertstr = '';\n"
                  . "    var invalid  = 0;\n\n";
@@ -799,11 +816,16 @@ sub render () {
         my $attr = $self->{fields}{$field} || {};
         debug 2, "$field: attr = $attr / attr->{value} = $attr->{value}";
 
+        # XHTML requires multiple="multiple". I just work here.
+        $attr->{multiple} &&= 'multiple';
+        delete $attr->{multiple} unless $attr->{multiple};
+        debug 1, "multiple = $attr->{multiple}" if $attr->{multiple};
+
         # print a label unless it's 0
         my $label = '';
         if ($args{labels} || ! exists $args{labels}) {
             $args{labels} = {} unless ref $args{labels} eq 'HASH';
-            $label = $attr->{label} || $args{labels}{$field} || _toname($field);
+            $label = _escapehtml($attr->{label} || $args{labels}{$field} || _toname($field));
             debug 1, "label for '$field' field set to '$label'";
         }
 
@@ -822,13 +844,13 @@ sub render () {
             debug 2, "$field: reset value to '@value' as src was CGI and sticky => 0";
         }
 
+        # set default field type to fieldtype if exists
+        $attr->{type} ||= $args{fieldtype} if $args{fieldtype};
+
         # Pre-catch setting type to 'static', by twiddling an extra flag, which
         # then tells our else statement below to also print out the value
         my $static = $attr->{type} eq 'static' || $args{static};
 
-        # set default field type to fieldtype if exists
-        $attr->{type} ||= $args{fieldtype} if $args{fieldtype};
-
         # in fact, allow a whole fieldattr thing to work globally
         if ($args{fieldattr} && ref $args{fieldattr} eq 'HASH') {
             while(my($k,$v) = each %{ $args{fieldattr} }) {
@@ -899,6 +921,7 @@ sub render () {
             $attr->{type} = 'hidden';
         }
 
+        $attr->{type} &&= lc $attr->{type};      # catch for type => 'TEXT'
         debug 1, "generating '$field' field as type '$attr->{type}' (@value)";
 
         # We now create the validation JavaScript, if it was requested
@@ -917,7 +940,7 @@ sub render () {
 
             # Touch a special element so that this element's label prints out bold
             # This is actually an HTML feature that must be nested here...
-            $self->{fields}{$field}{required} = 1;
+            $self->{fields}{$field}{required} = 1 if $need{$field};
 
             # Check our hash to see if it's a special pattern
             ($pattern, $helptag) = _data($VALID{$pattern}) if $VALID{$pattern};
@@ -945,6 +968,7 @@ sub render () {
                 my $et = $attr->{type} || 'text';
                 $et = 'input' if $et eq 'text';
                 my $alertstr = sprintf $args{messages}{"js_invalid_$et"}, $label;
+                $alertstr =~ s/'/\\'/g;     # handle embedded '
                 $alertstr .= '\n';
 
                 if ($attr->{type} eq 'select') {
@@ -1015,9 +1039,9 @@ EOF
 
                 if ($pattern =~ m#^m?(\S)(.*)\1$#) {
                     # JavaScript regexp
-                    ($pattern = $2) =~ s/\\\//\//g;
-                    $pattern =~ s/\//\\\//g;
-                    $jsfunc .= qq($in    if ($nn (! $jsfield.match(/$pattern/)) ) {\n);
+                    (my $tpat = $2) =~ s/\\\//\//g;
+                    $tpat =~ s/\//\\\//g;
+                    $jsfunc .= qq($in    if ($nn (! $jsfield.match(/$tpat/)) ) {\n);
                 } elsif (ref $pattern eq 'ARRAY') {
                     # must be w/i this set of values
                     # can you figure out how this piece of Perl works? ha ha ha ha ....
@@ -1045,7 +1069,9 @@ EOF
         HTMLGEN:
 
         # Save options for template creation below
-        my @opt = ();
+        # "options" are the options for our select list
+        my @opt = _opt($attr->{options} ||= $vattr);
+        @opt = _sort(@opt, $attr->{sortopts}) if $attr->{sortopts};
 
         # Must catch this to prevent non-existent fault below
         $attr->{labels} = {} unless ref $attr->{labels} eq 'HASH';
@@ -1056,64 +1082,71 @@ EOF
 
             $attr->{onChange} = delete $attr->{jsclick} if $attr->{jsclick};
 
-            # "options" are the options for our select list
-            @opt = _opt($attr->{options} ||= $vattr);
-            @opt = _sort(@opt, $attr->{sortopts}) if $attr->{sortopts};
+            # handle multiples.
+            $attr->{multiple} = 'multiple' if @value > 1;   # auto-detect
+
             unshift @opt, ['', $args{messages}{form_select_default}]
                 if $args{smartness} && ! $attr->{multiple};
 
-            # generate our select tag. handle multiples.
-            my $mult = $attr->{multiple} || (@value > 1) ? ' multiple' : '';
-            $tag = _tag("select$mult", name => $field, %{$attr});
+            # generate our select tag. 
+            delete $attr->{type};     # prevent <select type="select">
+            $tag = _tag('select', name => $field, %{$attr});
             for my $opt (@opt) {
                 # Since our data structure is a series of ['',''] things,
                 # we get the name from that. If not, then it's a list
                 # of regular old data that we _toname if nameopts => 1 
                 my($o,$n) = (ref $opt eq 'ARRAY') ? (@{$opt}) : ($opt);
                 $n ||= $attr->{labels}{$o} || ($attr->{nameopts} ? _toname($o) : $o);
-                my $slct = _ismember($o, @value) ? ' selected' : '';
-                $tag .= _tag("option$slct", value => $o) . $n . '</option>';
+                my @slct = _ismember($o, @value) ? (selected => 'selected') : ();
+                $tag .= _tag('option', value => $o, @slct) . _escapehtml($n) . '</option>';
             }
             $tag .= '</select>';
+            $attr->{type} = 'select';
 
         } elsif ($attr->{type} eq 'radio' || $attr->{type} eq 'checkbox') {
             $attr->{onClick} = delete $attr->{jsclick} if $attr->{jsclick};
 
+            # to allow checkboxes/radio buttons to wrap in columns
             my $checkbox_table = 0;  # toggle
             my $col_num = 0;
-            if ($attr->{type} eq 'checkbox' 
-                && defined($attr->{columns}) && $attr->{columns} > 0)
-            {
+            if (defined($attr->{columns}) && $attr->{columns} > 0) {
                 $checkbox_table = 1;
                 $tag .= _tag('table', %{$args{table}});
             }
 
-            # get our options
-            @opt = _opt($attr->{options} ||= $vattr);
-            @opt = _sort(@opt, $attr->{sortopts}) if $attr->{sortopts};
+            # for making an unordered list out of our options (for both checkbox and radio)
+            my $checkbox_ulist = 0;
+            if ($attr->{ulist}) {
+                $checkbox_ulist = 1;
+                $tag .= '<ul>';
+            }
 
             for my $opt (@opt) {
                 #  Divide up checkboxes in a user-controlled manner
                 if ($checkbox_table) {
-                    $tag .= "<tr>\n" if $col_num % $attr->{columns} == 0;
+                    $tag .= '<tr>' if $col_num % $attr->{columns} == 0;
                     $tag .= '<td>' . $font;
                 }
+                $tag .= '<li>' if $checkbox_ulist;
                 # Since our data structure is a series of ['',''] things,
                 # we get the name from that. If not, then it's a list
                 # of regular old data that we _toname if nameopts => 1 
                 my($o,$n) = (ref $opt eq 'ARRAY') ? (@{$opt}) : ($opt);
                 $n ||= $attr->{labels}{$o} || ($attr->{nameopts} ? _toname($o) : $o);
-                my $slct = _ismember($o, @value) ? ' checked' : '';
-                $tag .= _tag("input$slct", name => $field, value => $o, %{$attr}) 
-                      . ' ' . $n . ' ';
-                $tag .= '<br>' if $attr->{linebreaks};
-                $tag .= "</tr>\n" if $checkbox_table && $col_num++ % $attr->{columns} == 0;
+                my @slct = _ismember($o, @value) ? (checked => 'checked') : ();
+                $tag .= _tag('input', name => $field, value => $o, id => "${field}_$o", %{$attr}, @slct) 
+                      . ' ' . _tag('label', for => "${field}_$o") . _escapehtml($n) . '</label> ';
+                $tag .= '<br />' if $attr->{linebreaks};
+                $tag .= '</td></tr>' if $checkbox_table && 
+                                     ($col_num++ % $attr->{columns} == 0 || $col_num == @opt);
+                $tag .= "</li>\n" if $checkbox_ulist;
             }
-            $tag .= "</table>\n" if $checkbox_table;
+            $tag .= '</table>' if $checkbox_table;
+            $tag .= '</ul>' if $checkbox_ulist;
         } elsif ($attr->{type} eq 'textarea') {
             $attr->{onChange} = delete $attr->{jsclick} if $attr->{jsclick};
             my $text = join "\n", @value;
-            $tag = _tag('textarea', name => $field, %{$attr}) . _escapehtml($text) . "</textarea>";
+            $tag = _tag('textarea', name => $field, %{$attr}) . _escapehtml($text) . '</textarea>';
 
         } else {
             $attr->{onChange} = delete $attr->{jsclick} if $attr->{jsclick};
@@ -1135,10 +1168,25 @@ EOF
                 # Check for passwords
                 $value = '********' if $attr->{type_orig} eq 'password' && $args{smartness};
 
-                # print the value out too when in a static context
-                my $tagcom = _escapehtml($value);
-                $tag .= $tagcom . ' ' if $attr->{type} eq 'hidden' && $static && $tagcom;
-                $tag .= '<br>' if $attr->{linebreaks};
+                # This is a static field
+                if ($attr->{type} eq 'hidden' && $static) {
+                    # Lookup the description of the option, and print it
+                    # if available, otherwise it will use whatever value was set to.
+                    for my $opt (@opt) {
+                        # Since our data structure is a series of ['',''] things,
+                        # we get the name from that. If not, then it's a list
+                        # of regular old data that we _toname if nameopts => 1 
+                        my($o,$n) = (ref $opt eq 'ARRAY') ? (@{$opt}) : ($opt);
+
+                        if ($o eq $value) {
+                            # We found a matching option.
+                            $n ||= $attr->{labels}{$o} || ($attr->{nameopts} ? _toname($o) : $o);
+                            $tag .= _escapehtml($n) . ' ';
+                            last;
+                        }
+                    }
+                }
+                $tag .= '<br />' if $attr->{linebreaks};
             }
 
             # special catch to make life easier
@@ -1159,15 +1207,17 @@ EOF
         $et = 'input' if $et eq 'text';
         my $error = '';
         if (exists $self->{fields}{$field}{invalid}) {
-            $error = $self->{fields}{$field}{invalid}
-                  || $args{messages}{"form_invalid_$et"}
-                  || $args{messages}{form_invalid_field};
+            $error = $self->{fields}{$field}{invalid};  # custom message
+            if ($error eq 1) {    # hackish but effective
+                $error = $args{messages}{"form_invalid_$et"}
+                      || $args{messages}{form_invalid_field};
+            }
         }
         debug 2, "got error string = $error for form_invalid_$et";
 
         # if we have a template, then we setup the tag to a tmpl_var of the
         # same name via param(). otherwise, we generate HTML rows and stuff
-      # Added support for Text::Template and assigned type Text.
+        # Added support for Text::Template and assigned type Text.
         if (ref $args{template} eq 'HASH' && 
            ( $args{template}{type} eq 'TT2' || $args{template}{type} eq 'Text' ) ) {
             # Template Toolkit can access complex data pretty much unaided
@@ -1216,7 +1266,7 @@ EOF
                 my($slct, $chk) = _ismember($o, @value) ? ('selected', 'checked') : ('','');
                 debug 2, "<tmpl_loop loop-$field> = adding { label => $n, value => $o }";
                 push @tmpl_loop, {
-                    label => $n,
+                    label => _escapehtml($n),
                     value => $o,
                     checked => $chk,
                     selected => $slct,
@@ -1315,11 +1365,11 @@ EOJS
             if (ref $args{submit} eq 'ARRAY') {
                 # multiple buttons + JavaScript - here we go!
                 for my $s (_data $args{submit}) {
-                    my $js = ($args{submit} && $args{javascript})
-                                ? qq( onClick="this.form._submit.value = this.value;")
-                                : '';
-                    $submit .= _tag("input$js", type => 'submit', name => '_submit',
-                                value => $s);
+                    my @oncl = ($args{submit} && $args{javascript})
+                                ? (onClick => 'this.form._submit.value = this.value;')
+                                : ();
+                    $submit .= _tag('input', type => 'submit', name => '_submit',
+                                value => $s, @oncl);
                 }
             } else {
                 # show the text on the button
@@ -1467,23 +1517,32 @@ EOJS
             $outhtml = $header . $tt2output;
 
       } elsif( $tmpltype eq 'Text' ) {
-            # Text::Template support. Nearly identical to Template Toolkit support.
+            # Text::Template support. Similar to Template Toolkit support.
             eval { require Text::Template };
             puke "Can't use templates because Text::Template is not installed!" if $@;
+            # This sub taken helps us to support all of Text::Template's argument naming conventions
+            my $tt_param_name = sub {
+              my ($arg, %h) = @_;
+              my ($key) = grep { exists $h{$_} } ($arg, "\u$arg", "\U$arg", "-$arg", "-\u$arg", "-\U$arg");
+              return $key || $arg;
+            };
 
-            my ($tt2engine, $tt2data, $tt2var, $tt2output);
-            $tt2engine = $tmplopt{engine} || { }; 
-            unless (UNIVERSAL::isa($tt2engine, 'Text::Template')) {
-                $tt2engine->{TYPE}   ||= 'FILE';
-                $tt2engine->{SOURCE} ||= $tmplopt{template};
-                puke "Text::Template source not specified, use the 'template' option"
-                    unless $tt2engine->{SOURCE};
-                $tt2engine->{DELIMITERS} ||= [ '<%','%>' ];
-                $tt2engine = Text::Template->new(%$tt2engine)
+            my ($tt_engine, $tt_data, $tt_var, $tt_output, $tt_fill_in);
+            $tt_engine = $tmplopt{engine} || { }; 
+            unless (UNIVERSAL::isa($tt_engine, 'Text::Template')) {
+                $tt_engine->{&$tt_param_name('type',%$tt_engine)}   ||= 'FILE';
+                $tt_engine->{&$tt_param_name('source',%$tt_engine)} ||= $tmplopt{template} ||
+                    puke "Text::Template source not specified, use the 'template' option";
+                $tt_engine->{&$tt_param_name('delimiters',%$tt_engine)} ||= [ '<%','%>' ];
+                $tt_engine = Text::Template->new(%$tt_engine)
                     || puke $Text::Template::ERROR;
             }
-            $tt2data = $tmplopt{data} || {};
-            $tt2var  = $tmplopt{variable};      # optional var for nesting
+            if( ref($tmplopt{data}) eq 'ARRAY' ) {
+                $tt_data = $tmplopt{data};
+            } else {
+                $tt_data = [ $tmplopt{data} ];
+            }
+            $tt_var  = $tmplopt{variable};      # optional var for nesting
 
             # special fields
             $tmplvar{'title'}  = $args{title} || '';
@@ -1495,16 +1554,26 @@ EOJS
             $tmplvar{'invalid'} = $self->{state}{invalid};
             $tmplvar{'fields'} = [ map $tmplvar{field}{$_},
                    @{ $self->{field_names} } ];
-            if ($tt2var) {
-                $tt2data->{$tt2var} = \%tmplvar;
+            if ($tt_var) {
+                push @$tt_data, { $tt_var => \%tmplvar };
             } else {
-                $tt2data = { %$tt2data, %tmplvar };
+                push @$tt_data, \%tmplvar;
             }
 
-            $tt2output = $tt2engine->fill_in(HASH=> $tt2data)
+            $tt_fill_in = $tmplopt{fill_in} || {};
+            my $tt_fill_in_hash = $tt_fill_in->{&$tt_param_name('hash',%$tt_fill_in)} || {};
+            if( ref($tt_fill_in_hash) eq 'ARRAY' ) {
+                push @$tt_fill_in_hash, @$tt_data;
+            } else {
+                $tt_fill_in_hash = [ $tt_fill_in_hash, @$tt_data ];
+            }
+		
+            $tt_fill_in_hash = {} unless scalar(@$tt_fill_in_hash);
+            $tt_fill_in->{&$tt_param_name('hash',%$tt_fill_in)} = $tt_fill_in_hash;
+            $tt_output = $tt_engine->fill_in(%$tt_fill_in)
                 || puke "Text::Template expansion failed: $Text::Template::ERROR";
 
-            $outhtml = $header . $tt2output;
+            $outhtml = $header . $tt_output;
 
         } else {
             puke "Invalid template type '$tmpltype' specified - can be 'HTML' or 'TT2' or 'Text'";
@@ -1553,8 +1622,8 @@ sub mail () {
     my %args = _args(@_);
 
     # Where does the mailer live? Must be sendmail-compatible
-    my $mailer = '';
-    unless ($mailer = $args{mailer}) {
+    my $mailer = undef;
+    unless ($mailer = $args{mailer} && -x $mailer) {
         for my $sendmail (qw(/usr/lib/sendmail /usr/sbin/sendmail /usr/bin/sendmail)) {
             if (-x $sendmail) {
                 $mailer = "$sendmail -t";
@@ -1562,8 +1631,12 @@ sub mail () {
             }
         }
     }
-    unless ($mailer && -x $mailer) {
-        belch "Cannot find a sendmail-compatible mailer to use";
+    unless ($mailer) {
+        belch "Cannot find a sendmail-compatible mailer to use; mail aborting";
+        return;
+    }
+    unless ($args{to}) {
+        belch "Missing required 'to' argument; cannot continue without recipient";
         return;
     }
 
@@ -1571,15 +1644,13 @@ sub mail () {
     my $oldpath = $ENV{PATH};
     $ENV{PATH} = '/usr/bin:/usr/sbin';
 
-    open(MAIL, "|$mailer") || next;
-    print MAIL <<EOF;
-From: $args{from}
-To: $args{to}
-Cc: $args{cc}
-Subject: $args{subject}
+    open(MAIL, "|$mailer >/dev/null 2>&1") || next;
+    print MAIL "From: $args{from}\n";
+    print MAIL "To: $args{to}\n";
+    print MAIL "Cc: $args{cc}\n" if $args{cc};
+    print MAIL "Subject: $args{subject}\n\n";
+    print MAIL "$args{text}\n";
 
-$args{text}
-EOF
     # retaint
     $ENV{PATH} = $oldpath;
 
@@ -1624,8 +1695,18 @@ sub mailresults () {
     # subject default
     $args{subject} ||= "$self->{opt}{title} Submission Results";
 
+    if ($args{skip}) {
+        if ($args{skip} =~ m#^m?(\S)(.*)\1$#) {
+            ($args{skip} = $2) =~ s/\\\//\//g;
+            $args{skip} =~ s/\//\\\//g;
+        }
+    }
+
     my @form = ();
     for my $field ($self->fields) {
+        if ($args{skip} && $field =~ /$args{skip}/) {
+            next;
+        }
         my $v = join $join, $self->field($field);
         $field = _toname($field) if $args{labels};
         push @form, "$field$delim$v"; 
@@ -1739,32 +1820,32 @@ sub validate () {
 
             if ($pattern =~ m#^m?(\S)(.*)\1$#) {
                 # it be a regexp
-                ($pattern = $2) =~ s/\\\//\//g;
-                $pattern =~ s/\//\\\//g;
-                debug 1, "$field: does '$value' =~ /$pattern/ ?";
-                unless ($value =~ /$pattern/) {
-                    $self->{fields}{$field}{invalid} = undef;
+                (my $tpat = $2) =~ s/\\\//\//g;
+                $tpat =~ s/\//\\\//g;
+                debug 1, "$field: does '$value' =~ /$tpat/ ?";
+                unless ($value =~ /$tpat/) {
+                    $self->{fields}{$field}{invalid} = 1;
                     $thisfail = ++$bad;
                 }
             } elsif (ref $pattern eq 'ARRAY') {
                 # must be w/i this set of values
                 debug 1, "$field: is '$value' in (@{$pattern}) ?";
                 unless (_ismember($value, @{$pattern})) {
-                    $self->{fields}{$field}{invalid} = undef;
+                    $self->{fields}{$field}{invalid} = 1;
                     $thisfail = ++$bad;
                 }
             } elsif (ref $pattern eq 'CODE') {
                 # eval that mofo, which gives them $form
                 debug 1, "$field: does $pattern($value) ret true ?";
                 unless ( &{$pattern}($value) ) {
-                    $self->{fields}{$field}{invalid} = undef;
+                    $self->{fields}{$field}{invalid} = 1;
                     $thisfail = ++$bad;
                 }
             } elsif ($pattern eq 'VALUE') {
                 # Not null
                 debug 1, "$field: length '$value' > 0 ?";
                 unless (defined($value) && length($value)) {
-                    $self->{fields}{$field}{invalid} = undef;
+                    $self->{fields}{$field}{invalid} = 1;
                     $thisfail = ++$bad;
                 }
             } else {
@@ -1774,7 +1855,7 @@ sub validate () {
                 # must escape to prevent serious problem if $value = "'; system 'rm -f /'; '"
                 debug 1, "$field: '$value' $pattern ? 1 : 0";
                 unless (eval qq(\$value $pattern ? 1 : 0)) {
-                    $self->{fields}{$field}{invalid} = undef;
+                    $self->{fields}{$field}{invalid} = 1;
                     $thisfail = ++$bad;
                 }
                 belch "Literal code eval error in validate: $@" if $@;
@@ -1786,7 +1867,7 @@ sub validate () {
         }
         # If not $atleastone and they asked for validation, then we
         # know that we have an error since this means no values
-        unless ($atleastone) { $self->{fields}{$field}{invalid} = undef; $bad++; }
+        unless ($atleastone) { $self->{fields}{$field}{invalid} = 1; $bad++; }
     }
     debug 2, "validation done, failures (\$bad) = $bad";
     $self->{state}{invalid} = $bad;
